@@ -7,7 +7,7 @@ import os
 import logging
 import traceback
 
-from datadog_lambda.cold_start import set_cold_start
+from datadog_lambda.cold_start import set_cold_start, is_cold_start
 from datadog_lambda.metric import (
     lambda_stats,
     submit_invocations_metric,
@@ -16,11 +16,13 @@ from datadog_lambda.metric import (
 from datadog_lambda.patch import patch_all
 from datadog_lambda.tracing import (
     extract_dd_trace_context,
-    set_correlation_ids,
     inject_correlation_ids,
+    dd_tracing_enabled,
+    set_correlation_ids,
+    set_dd_trace_py_root,
+    create_function_execution_span,
 )
 
-
 logger = logging.getLogger(__name__)
 
 
@@ -39,7 +41,6 @@ def my_lambda_handle(event, context):
 
 
 class _NoopDecorator(object):
-
     def __init__(self, func):
         self.func = func
 
@@ -53,37 +54,50 @@ class _LambdaDecorator(object):
     and extracts/injects trace context.
     """
 
-    _force_new = False
+    _force_wrap = False
 
     def __new__(cls, func):
         """
         If the decorator is accidentally applied to the same function multiple times,
-        only the first one takes effect.
+        wrap only once.
 
-        If _force_new, always return a real decorator, useful for unit tests.
+        If _force_wrap, always return a real decorator, useful for unit tests.
         """
-        if cls._force_new or not getattr(func, "_dd_wrapped", False):
-            wrapped = super(_LambdaDecorator, cls).__new__(cls)
-            wrapped._dd_wrapped = True
-            return wrapped
-        else:
-            return _NoopDecorator(func)
+        try:
+            if cls._force_wrap or not isinstance(func, _LambdaDecorator):
+                wrapped = super(_LambdaDecorator, cls).__new__(cls)
+                logger.debug("datadog_lambda_wrapper wrapped")
+                return wrapped
+            else:
+                logger.debug("datadog_lambda_wrapper already wrapped")
+                return _NoopDecorator(func)
+        except Exception:
+            traceback.print_exc()
+            return func
 
     def __init__(self, func):
         """Executes when the wrapped function gets wrapped"""
-        self.func = func
-        self.flush_to_log = os.environ.get("DD_FLUSH_TO_LOG", "").lower() == "true"
-        self.logs_injection = (
-            os.environ.get("DD_LOGS_INJECTION", "true").lower() == "true"
-        )
+        try:
+            self.func = func
+            self.flush_to_log = os.environ.get("DD_FLUSH_TO_LOG", "").lower() == "true"
+            self.logs_injection = (
+                os.environ.get("DD_LOGS_INJECTION", "true").lower() == "true"
+            )
+            self.merge_xray_traces = (
+                os.environ.get("DD_MERGE_XRAY_TRACES", "false").lower() == "true"
+            )
+            self.function_name = os.environ.get("AWS_LAMBDA_FUNCTION_NAME", "function")
 
-        # Inject trace correlation ids to logs
-        if self.logs_injection:
-            inject_correlation_ids()
+            # Inject trace correlation ids to logs
+            if self.logs_injection:
+                inject_correlation_ids()
 
-        # Patch HTTP clients to propagate Datadog trace context
-        patch_all()
-        logger.debug("datadog_lambda_wrapper initialized")
+            # Patch third-party libraries for tracing
+            patch_all()
+
+            logger.debug("datadog_lambda_wrapper initialized")
+        except Exception:
+            traceback.print_exc()
 
     def __call__(self, event, context, **kwargs):
         """Executes when the wrapped function gets called"""
@@ -97,14 +111,27 @@ class _LambdaDecorator(object):
             self._after(event, context)
 
     def _before(self, event, context):
-        set_cold_start()
         try:
+
+            set_cold_start()
             submit_invocations_metric(context)
             # Extract Datadog trace context from incoming requests
-            extract_dd_trace_context(event)
+            dd_context = extract_dd_trace_context(event)
 
-            # Set log correlation ids using extracted trace context
-            set_correlation_ids()
+            self.span = None
+            if dd_tracing_enabled:
+                set_dd_trace_py_root(dd_context, self.merge_xray_traces)
+                self.span = create_function_execution_span(
+                    context,
+                    self.function_name,
+                    is_cold_start(),
+                    dd_context,
+                    self.merge_xray_traces,
+                )
+            else:
+                set_correlation_ids()
+
+            logger.debug("datadog_lambda_wrapper _before() done")
         except Exception:
             traceback.print_exc()
 
@@ -112,6 +139,9 @@ class _LambdaDecorator(object):
         try:
             if not self.flush_to_log:
                 lambda_stats.flush(float("inf"))
+            if self.span:
+                self.span.finish()
+            logger.debug("datadog_lambda_wrapper _after() done")
         except Exception:
             traceback.print_exc()
 
